// src/parser/FunctionManager.js
const discord = require('discord.js');
const ms = require('ms');

function safeJsonParse(str) {
    try {
        return JSON.parse(str);
    } catch {
        return null;
    }
}

class FunctionManager {
    constructor(botInstance, parserInstance) {
        this.bot = botInstance;
        this.parser = parserInstance;
        this.functions = new Map();
        this._registerFunctions();
    }

    async execute(name, args, context) {
        const funcData = this.functions.get(name.toLowerCase());

        if (!funcData) {
            const letValue = context.variables?.[name.toLowerCase()];
            if (letValue !== undefined) return letValue;
            const contextVar = this._getContextVariable(name, context);
            if (contextVar !== undefined) return contextVar;
            throw new Error(`Parser Error: Unknown function or variable "$${name}"`);
        }

        if (args.length < funcData.minArgs || (funcData.maxArgs !== Infinity && args.length > funcData.maxArgs)) {
           throw new Error(`Parser Error: Invalid argument count for $${name}. Expected ${funcData.minArgs}${funcData.maxArgs === funcData.minArgs ? '' : funcData.maxArgs === Infinity ? '+' : '-' + funcData.maxArgs}, got ${args.length}.`);
        }

        try {
            return await funcData.execute(args, context);
        } catch (error) {
            console.error(`[Relay.js Func Error] Error executing $${name}:`, error);
            throw new Error(`Error in $${name}: ${error.message}`);
        }
    }

     _getContextVariable(name, context) {
         const msg = context.message;
         const interaction = context.interaction;
         const user = interaction?.user || msg?.author;
         const member = interaction?.member || msg?.member;
         const channel = interaction?.channel || msg?.channel;
         const guild = interaction?.guild || msg?.guild;
         const interactionOptions = context.interaction?.options;

         switch (name.toLowerCase()) {
             case 'messagecontent': return msg?.content;
             case 'messageid': return msg?.id;
             case 'args': return msg?.content?.split(' ').slice(1).join(' ');
             case 'argscount': return msg?.content?.split(' ').slice(1).filter(Boolean).length ?? 0;
             case 'interactionid': return interaction?.id;
             case 'customid': return interaction?.customId;
             case 'commandname': return interaction?.commandName;
             case 'selectedvalues': return context.selectedValues?.join(';');
             case 'authorid': case 'userid': return user?.id;
             case 'authortag': case 'usertag': return user?.tag;
             case 'authormention': case 'usermention': return user?.toString();
             case 'authorname': case 'username': return user?.username;
             case 'useravatar': return user?.displayAvatarURL({ dynamic: true, size: 1024 });
             case 'memberjoineddate': return member?.joinedTimestamp;
             case 'accountcreationdate': return user?.createdTimestamp;
             case 'memberpermissions': return member?.permissions.toArray().join('; ');
             case 'usertimeout': return member?.communicationDisabledUntilTimestamp;
             case 'channelid': return channel?.id;
             case 'channelname': return channel?.name;
             case 'channeltopic': return channel?.topic;
             case 'channelmention': return channel?.toString();
             case 'channeltype': return channel ? discord.ChannelType[channel.type] : undefined;
             case 'guildid': return guild?.id;
             case 'guildname': return guild?.name;
             case 'guildicon': return guild?.iconURL({ dynamic: true, size: 1024 });
             case 'membercount': return guild?.memberCount;
             case 'ownerid': return guild?.ownerId;
             case 'botid': return this.bot.client.user?.id;
             case 'bottag': return this.bot.client.user?.tag;
             case 'botmention': return this.bot.client.user?.toString();
             case 'botping': return this.bot.client.ws.ping;
             case 'timestamp': return Date.now();
             default:
                if (name.toLowerCase().startsWith('option_') && interactionOptions) {
                    const optionName = name.substring('option_'.length);
                    return interactionOptions.get(optionName)?.value;
                }
                if (name.toLowerCase().startsWith('userid_') && interactionOptions) {
                    const optionName = name.substring('userid_'.length);
                    return interactionOptions.getUser(optionName)?.id;
                }
                 if (name.toLowerCase().startsWith('roleid_') && interactionOptions) {
                    const optionName = name.substring('roleid_'.length);
                    return interactionOptions.getRole(optionName)?.id;
                }
                 if (name.toLowerCase().startsWith('channelid_') && interactionOptions) {
                    const optionName = name.substring('channelid_'.length);
                    return interactionOptions.getChannel(optionName)?.id;
                }
                return context.variables?.[name.toLowerCase()];
         }
     }

    async _getMember(userId, context) {
        const guild = context.guild;
        if (!guild || !userId) return null;
        try {
            return await guild.members.fetch(userId);
        } catch {
            return null;
        }
    }

     _parseDuration(durationStr) {
        if (!durationStr) return null;
        try {
            const milliseconds = ms(String(durationStr));
            return milliseconds === undefined ? null : milliseconds;
        } catch {
            return null;
        }
    }

    _registerFunctions() {
        const register = (name, executeFunc, minArgs = 0, maxArgs = Infinity) => {
            this.functions.set(name.toLowerCase(), { execute: executeFunc, minArgs, maxArgs });
        };

        register('reply', async (args, context) => {
            const options = this._parseMessageOptions(args[0]);
            if (!options.content && !options.embeds && !options.components) throw new Error("Cannot send an empty reply.");
            options.ephemeral = String(args[1]).toLowerCase() === 'true' || String(args[1]).toLowerCase() === 'ephemeral';
            options.allowedMentions = { repliedUser: false, ...options.allowedMentions };

            if (context.message) {
                try { return (await context.message.reply(options)).id; }
                catch (e) { throw new Error(`Failed to reply to message: ${e.message}`); }
            } else if (context.interaction) {
                try {
                    if (context.interaction.replied || context.interaction.deferred) {
                        return (await context.interaction.followUp(options)).id;
                    } else {
                        await context.interaction.reply(options);
                         try { return (await context.interaction.fetchReply()).id; } catch { return context.interaction.id; }
                    }
                } catch (e) { throw new Error(`Failed to reply to interaction: ${e.message}`); }
            } else {
                throw new Error("$reply requires message or interaction context.");
            }
        }, 1, 2);

        register('sendMessage', async (args, context) => {
            const channelId = String(args[0]);
            const options = this._parseMessageOptions(args[1]);
            if (!channelId) throw new Error("Channel ID is required for $sendMessage.");
            if (!options.content && !options.embeds && !options.components) throw new Error("Cannot send an empty message.");
            const channel = await this.bot.client.channels.fetch(channelId).catch(() => null);
            if (!channel || !channel.isTextBased()) throw new Error(`Invalid or non-text channel ID provided: ${channelId}`);
            try {
                 if (args[2]) {
                     options.reply = { messageReference: String(args[2]), failIfNotExists: false };
                 }
                return (await channel.send(options)).id;
            } catch (e) {
                throw new Error(`Failed to send message to channel ${channelId}: ${e.message}`);
            }
        }, 2, 3);

        register('log', (args, context) => {
            console.log("[Relay.js Log]", ...args.map(String));
            return '';
        }, 1, Infinity);

        register('if', (args, context) => {
             if (args.length < 2 || args.length > 3) throw new Error("Usage: $if[condition;then;else?]");
             const condition = args[0];
             const thenBranch = args[1];
             const elseBranch = args[2];
             const isTruthy = condition !== 'false' && condition !== '0' && condition !== '' && condition !== false && condition !== 0 && condition != null;
             return isTruthy ? thenBranch : (elseBranch ?? '');
         }, 2, 3);

         register('stop', (args, context) => {
            context.stopExecution = true;
            return '';
         }, 0, 0);

         register('stopIf', (args, context) => {
             if (args.length !== 1) throw new Error("Usage: $stopIf[condition]");
             const condition = args[0];
             const isTruthy = condition !== 'false' && condition !== '0' && condition !== '' && condition !== false && condition !== 0 && condition != null;
             if (isTruthy) context.stopExecution = true;
             return '';
         }, 1, 1);

        register('let', (args, context) => {
            if (args.length !== 2) throw new Error("Usage: $let[variableName;value]");
            const varName = String(args[0]).toLowerCase();
            if (!varName.match(/^[\w-]+$/)) throw new Error("Invalid variable name.");
            if (!context.variables) context.variables = {};
            context.variables[varName] = args[1];
            return '';
        }, 2, 2);

        register('get', (args, context) => {
             if (args.length !== 1) throw new Error("Usage: $get[variableName]");
             const varName = String(args[0]).toLowerCase();
             return context.variables?.[varName] ?? '';
         }, 1, 1);

         register('ping', (args, context) => this.bot.client.ws.ping, 0, 0);
         register('args', (args, context) => context.message?.content?.split(' ').slice(1).join(' ') ?? '', 0, 0);
         register('arg', (args, context) => {
             const index = parseInt(args[0], 10);
             if (isNaN(index) || index < 1) throw new Error("Usage: $arg[index] - index must be 1 or higher.");
             const messageArgs = context.message?.content?.split(' ').slice(1);
             return messageArgs?.[index - 1] ?? '';
         }, 1, 1);
        register('isEmpty', (args) => (args[0] === null || args[0] === undefined || String(args[0]).trim() === ''), 1, 1);
         register('ignoreBots', (args, context) => {
             if (context.message?.author?.bot) context.stopExecution = true;
             return '';
         }, 0, 0);
         register('onlyPerms', async (args, context) => {
            if (args.length < 1) throw new Error("Usage: $onlyPerms[perm1;perm2;...;optionalReplyMessage]");
             const member = context.interaction?.member || context.message?.member;
             const defaultReply = "You do not have the required permissions!";
             let replyMessage = defaultReply;
             let requiredPerms = [...args];
             if (args.length > 1 && typeof args[args.length - 1] === 'string' && !discord.PermissionsBitField.Flags[args[args.length - 1]]) {
                 replyMessage = args[args.length - 1];
                 requiredPerms.pop();
             }
             if (!member) {
                 context.stopExecution = true;
                 await this.parser.sendErrorReply(new Error(replyMessage === defaultReply ? "Permission check failed (no server context)." : replyMessage), context);
                 return '';
             }
             const missingPerms = [];
             for (const permName of requiredPerms) {
                if (typeof permName !== 'string') continue;
                 const permBit = discord.PermissionsBitField.Flags[permName];
                 if (permBit === undefined) {
                     console.warn(`[Relay.js Warn] Invalid permission name used in $onlyPerms: ${permName}`);
                     continue;
                 }
                 if (!member.permissions.has(permBit)) {
                     missingPerms.push(permName);
                 }
             }
             if (missingPerms.length > 0) {
                 context.stopExecution = true;
                 const finalReply = `${replyMessage} (Missing: ${missingPerms.join(', ')})`;
                  await this.parser.sendErrorReply(new Error(finalReply), context);
                 return '';
             }
             return '';
         }, 1, Infinity);

         register('option', (args, context) => {
             if (!context.interaction?.isChatInputCommand()) return '';
             if (args.length !== 1) throw new Error("Usage: $option[optionName]");
             const optionName = String(args[0]).toLowerCase();
             const optionsResolver = context.interaction.options;
             const optionData = optionsResolver.get(optionName);
             if (!optionData) return '';
             switch (optionData.type) {
                 case discord.ApplicationCommandOptionType.Attachment: return optionData.attachment?.url ?? '';
                 case discord.ApplicationCommandOptionType.User: return optionData.user?.id ?? '';
                 case discord.ApplicationCommandOptionType.Channel: return optionData.channel?.id ?? '';
                 case discord.ApplicationCommandOptionType.Role: return optionData.role?.id ?? '';
                 case discord.ApplicationCommandOptionType.Mentionable: return optionData.member?.id || optionData.user?.id || optionData.role?.id || '';
                 default: return optionData.value ?? '';
             }
         }, 1, 1);

         register('kick', async (args, context) => {
             await context.bot.parser.functions.execute('onlyPerms', ['KickMembers'], context);
             if (context.stopExecution) return '';
             const userId = String(args[0]);
             const reason = String(args[1] ?? 'No reason specified.');
             if (!userId) throw new Error("Usage: $kick[userID; reason?]");
             if (!context.guild) throw new Error("$kick requires a guild context.");
             const member = await this._getMember(userId, context);
             if (!member) throw new Error(`User with ID ${userId} not found in this guild.`);
             if (!member.kickable) throw new Error(`I do not have permission to kick ${member.user.tag}.`);
             try {
                 await member.kick(reason);
                 return 'true';
             } catch (e) {
                 throw new Error(`Failed to kick ${member.user.tag}: ${e.message}`);
             }
         }, 1, 2);

         register('ban', async (args, context) => {
             await context.bot.parser.functions.execute('onlyPerms', ['BanMembers'], context);
             if (context.stopExecution) return '';
             const userId = String(args[0]);
             const reason = String(args[1] ?? 'No reason specified.');
             const deleteDays = parseInt(args[2], 10);
             const deleteMessageSeconds = isNaN(deleteDays) ? 0 : Math.max(0, Math.min(deleteDays, 7)) * 86400;
             if (!userId) throw new Error("Usage: $ban[userID; reason?; deleteDays?]");
             if (!context.guild) throw new Error("$ban requires a guild context.");
             try {
                 await context.guild.members.ban(userId, { reason: reason, deleteMessageSeconds: deleteMessageSeconds });
                 return 'true';
             } catch (e) {
                 const member = await this._getMember(userId, context).catch(() => null);
                 if (member && !member.bannable) {
                     throw new Error(`I do not have permission to ban ${member.user.tag}.`);
                 }
                 throw new Error(`Failed to ban user ${userId}: ${e.message}`);
             }
         }, 1, 3);

         register('timeout', async (args, context) => {
            await context.bot.parser.functions.execute('onlyPerms', ['ModerateMembers'], context);
            if (context.stopExecution) return '';
             const userId = String(args[0]);
             const durationStr = String(args[1]);
             const reason = String(args[2] ?? 'No reason specified.');
             if (!userId || !durationStr) throw new Error("Usage: $timeout[userID; duration; reason?]");
             if (!context.guild) throw new Error("$timeout requires a guild context.");
             const member = await this._getMember(userId, context);
             if (!member) throw new Error(`User with ID ${userId} not found in this guild.`);
             if (!member.moderatable) throw new Error(`I do not have permission to timeout ${member.user.tag}.`);
             const durationMs = this._parseDuration(durationStr);
             if (durationMs === null || durationMs <= 0) {
                 if (durationStr === '0' || durationStr.toLowerCase() === 'remove' || durationStr.toLowerCase() === 'none') {
                      try {
                         await member.timeout(null, reason);
                         return 'true';
                      } catch (e) {
                         throw new Error(`Failed to remove timeout for ${member.user.tag}: ${e.message}`);
                      }
                 } else {
                     throw new Error(`Invalid duration format: "${durationStr}". Use formats like '5m', '1h', '1d 12h'. Use '0' or 'remove' to clear timeout.`);
                 }
             }
              if (durationMs > 28 * 24 * 60 * 60 * 1000) {
                  throw new Error("Timeout duration cannot exceed 28 days.");
              }
             try {
                 await member.timeout(durationMs, reason);
                 return 'true';
             } catch (e) {
                 throw new Error(`Failed to timeout ${member.user.tag}: ${e.message}`);
             }
         }, 2, 3);

          register('addRole', async (args, context) => {
            await context.bot.parser.functions.execute('onlyPerms', ['ManageRoles'], context);
            if (context.stopExecution) return '';
             const userId = String(args[0]);
             const roleId = String(args[1]);
             const reason = String(args[2] ?? 'No reason specified.');
             if (!userId || !roleId) throw new Error("Usage: $addRole[userID; roleID; reason?]");
             if (!context.guild) throw new Error("$addRole requires a guild context.");
             const member = await this._getMember(userId, context);
             if (!member) throw new Error(`User with ID ${userId} not found in this guild.`);
             const role = await context.guild.roles.fetch(roleId).catch(() => null);
             if (!role) throw new Error(`Role with ID ${roleId} not found.`);
             if (role.managed) throw new Error(`Cannot add managed role "${role.name}".`);
             if (context.guild.members.me?.roles.highest.comparePositionTo(role) <= 0) {
                 throw new Error(`My highest role is not high enough to add role "${role.name}".`);
             }
             try {
                 await member.roles.add(role, reason);
                 return 'true';
             } catch (e) {
                 throw new Error(`Failed to add role "${role.name}" to ${member.user.tag}: ${e.message}`);
             }
         }, 2, 3);

         register('removeRole', async (args, context) => {
             await context.bot.parser.functions.execute('onlyPerms', ['ManageRoles'], context);
             if (context.stopExecution) return '';
             const userId = String(args[0]);
             const roleId = String(args[1]);
             const reason = String(args[2] ?? 'No reason specified.');
             if (!userId || !roleId) throw new Error("Usage: $removeRole[userID; roleID; reason?]");
             if (!context.guild) throw new Error("$removeRole requires a guild context.");
             const member = await this._getMember(userId, context);
             if (!member) throw new Error(`User with ID ${userId} not found in this guild.`);
             const role = await context.guild.roles.fetch(roleId).catch(() => null);
             if (!role) throw new Error(`Role with ID ${roleId} not found.`);
             if (role.managed) throw new Error(`Cannot remove managed role "${role.name}".`);
             if (context.guild.members.me?.roles.highest.comparePositionTo(role) <= 0) {
                 throw new Error(`My highest role is not high enough to remove role "${role.name}".`);
             }
             if (!member.roles.cache.has(role.id)) return 'false';
             try {
                 await member.roles.remove(role, reason);
                 return 'true';
             } catch (e) {
                 throw new Error(`Failed to remove role "${role.name}" from ${member.user.tag}: ${e.message}`);
             }
         }, 2, 3);

         register('hasRole', async (args, context) => {
             const userId = String(args[0]);
             const roleId = String(args[1]);
             if (!userId || !roleId) throw new Error("Usage: $hasRole[userID; roleID]");
             if (!context.guild) return 'false';
             const member = await this._getMember(userId, context);
             return member?.roles.cache.has(roleId) ? 'true' : 'false';
         }, 2, 2);

        register('random', (args) => {
            const min = parseInt(args[0], 10);
            const max = parseInt(args[1], 10);
            if (isNaN(min) || isNaN(max)) throw new Error("Usage: $random[min; max] - requires numbers.");
            if (min > max) throw new Error("Min cannot be greater than max in $random.");
            return Math.floor(Math.random() * (max - min + 1)) + min;
        }, 2, 2);

        register('randomText', (args) => {
            if (args.length === 0) return '';
            const randomIndex = Math.floor(Math.random() * args.length);
            return args[randomIndex];
        }, 1, Infinity);

        register('formatDate', (args) => {
            const timestamp = parseInt(args[0], 10);
            const format = String(args[1] || 'f');
            if (isNaN(timestamp)) throw new Error("Usage: $formatDate[unixTimestampMs; format?] - requires valid timestamp.");
            const validFormats = ['t', 'T', 'd', 'D', 'f', 'F', 'R'];
            if (!validFormats.includes(format)) throw new Error(`Invalid format "${format}". Use: t, T, d, D, f, F, R.`);
            return `<t:${Math.floor(timestamp / 1000)}:${format}>`;
        }, 1, 2);

        register('toUpperCase', (args) => String(args[0] ?? '').toUpperCase(), 1, 1);
        register('toLowerCase', (args) => String(args[0] ?? '').toLowerCase(), 1, 1);
        register('replaceText', (args) => {
            if (args.length !== 3) throw new Error("Usage: $replaceText[text; searchFor; replaceWith]");
            return String(args[0]).replace(new RegExp(String(args[1]), 'g'), String(args[2]));
        }, 3, 3);
        register('textLength', (args) => String(args[0] ?? '').length, 1, 1);
        register('substring', (args) => {
            const text = String(args[0] ?? '');
            const start = parseInt(args[1], 10);
            const end = args[2] !== undefined ? parseInt(args[2], 10) : undefined;
            if (isNaN(start)) throw new Error("Usage: $substring[text; startIndex; endIndex?] - startIndex required.");
            if (end !== undefined && isNaN(end)) throw new Error("EndIndex must be a number if provided.");
            return text.substring(start, end);
        }, 2, 3);

        register('sum', (args) => args.reduce((acc, val) => acc + (parseFloat(val) || 0), 0), 1, Infinity);
        register('subtract', (args) => {
             if (args.length !== 2) throw new Error("Usage: $subtract[num1; num2]");
             return (parseFloat(args[0]) || 0) - (parseFloat(args[1]) || 0);
         }, 2, 2);
        register('multiply', (args) => args.reduce((acc, val) => acc * (parseFloat(val) || 1), 1), 1, Infinity); // Start with 1 for multiplication
        register('divide', (args) => {
             if (args.length !== 2) throw new Error("Usage: $divide[num1; num2]");
             const divisor = parseFloat(args[1]) || 0;
             if (divisor === 0) throw new Error("Cannot divide by zero.");
             return (parseFloat(args[0]) || 0) / divisor;
         }, 2, 2);

        register('embed', (args) => {
            const structureString = String(args[0] ?? '{}');
             let embedData;
             try {
                 embedData = JSON.parse(structureString);
             } catch (e) {
                 throw new Error(`Invalid structure passed to $embed. Expected JSON-like object string. Error: ${e.message}`);
             }
             return JSON.stringify(embedData);
         }, 1, 1);

         register('button', (args) => {
             const customId = String(args[0]);
             const label = String(args[1]);
             const styleStr = String(args[2]).toLowerCase();
             const disabled = String(args[3]).toLowerCase() === 'true';
             const emoji = args[4] ? String(args[4]) : undefined;
             const url = args[5] ? String(args[5]) : undefined;
             if (!customId && styleStr !== 'link') throw new Error("Usage: $button[customID; label; style; disabled?; emoji?] - customID required unless style is 'link'.");
             if (!label && !emoji) throw new Error("Button requires a label or an emoji.");
              if (styleStr === 'link' && !url) throw new Error("Link buttons require a URL as the 6th argument.");
              if (styleStr !== 'link' && url) throw new Error("URL argument is only valid for link buttons.");
             const styleMap = { primary: 1, secondary: 2, success: 3, danger: 4, link: 5 };
             const style = styleMap[styleStr];
             if (style === undefined) throw new Error(`Invalid button style "${styleStr}". Use: primary, secondary, success, danger, link.`);
             const buttonData = {
                 type: 2, // ComponentType.Button
                 style: style,
                 label: label || undefined,
                 emoji: emoji,
                 disabled: disabled,
                 custom_id: style === 5 ? undefined : customId, // ButtonStyle.Link
                 url: style === 5 ? url : undefined
             };
             return JSON.stringify(buttonData);
         }, 3, 6);

         register('actionRow', (args) => {
            if (args.length === 0 || args.length > 5) throw new Error("$actionRow can contain 1 to 5 components.");
             const components = args.map(arg => {
                 const parsed = safeJsonParse(String(arg));
                 if (!parsed || typeof parsed !== 'object') throw new Error(`Invalid component JSON passed to $actionRow: ${String(arg).substring(0, 50)}...`);
                 return parsed;
             });
             const rowData = { type: 1, components: components }; // ComponentType.ActionRow
             return JSON.stringify(rowData);
         }, 1, 5);
    }

    _parseMessageOptions(inputArg) {
        const options = { content: null, embeds: [], components: [] };
        if (typeof inputArg === 'string') {
            const parsedJson = safeJsonParse(inputArg);
            if (parsedJson && typeof parsedJson === 'object') {
                if ('content' in parsedJson || 'embeds' in parsedJson || 'components' in parsedJson) {
                     options.content = parsedJson.content ? String(parsedJson.content) : null;
                     options.embeds = Array.isArray(parsedJson.embeds) ? parsedJson.embeds : [];
                     options.components = Array.isArray(parsedJson.components) ? parsedJson.components : [];
                }
                 else if ('title' in parsedJson || 'description' in parsedJson || 'fields' in parsedJson || 'color' in parsedJson) {
                     options.embeds.push(parsedJson);
                 }
                 else { options.content = inputArg; }
            } else { options.content = inputArg; }
        } else if (inputArg !== null && inputArg !== undefined) { options.content = String(inputArg); }

         if (options.embeds.length === 0) delete options.embeds;
         if (options.components.length === 0) delete options.components;
         if (!options.content && !options.embeds && !options.components) { options.content = null; } // Allow fully empty if needed
         else if (options.content === '') { options.content = null; }
        return options;
    }
}

module.exports = FunctionManager;