// src/parser/Parser.js
const FunctionManager = require('./FunctionManager');

class Parser {
    constructor(botInstance) {
        this.bot = botInstance;
        this.functions = new FunctionManager(botInstance, this);
    }

    async parse(script, context) {
        if (!script || typeof script !== 'string') return null;
        context.variables = context.variables || {};
        context.stopExecution = false;

        try {
            const substitutedScript = this.substituteContextVariables(script, context);
            const result = await this.evaluate(substitutedScript, context);
            if (context.stopExecution) return null;
            return result;
        } catch (error) {
            console.error("[Relay.js Parser Error] Error during script execution:", error);
            await this.sendErrorReply(error, context);
            return null;
        }
    }

     substituteContextVariables(script, context) {
         return script.replace(/(?<!\\)\$([\w-]+)(?!\[)/g, (match, varName) => {
             const value = this.functions._getContextVariable(varName, context);
             if (value !== undefined) {
                 return String(value);
             }
              const letValue = context.variables?.[varName.toLowerCase()];
              if (letValue !== undefined) {
                  return String(letValue);
              }
             return match;
         });
     }

    async evaluate(text, context) {
        if (context.stopExecution) return null;
        let result = '';
        let currentPos = 0;

        while (currentPos < text.length) {
            if (context.stopExecution) break;

            let funcStartIndex = -1;
            let searchPos = currentPos;
            while (searchPos < text.length) {
                const dollarIndex = text.indexOf('$', searchPos);
                if (dollarIndex === -1) break;

                let isDollarEscaped = false;
                if (dollarIndex > 0 && text[dollarIndex - 1] === '\\') {
                     let backslashCount = 0;
                     let k = dollarIndex - 1;
                     while (k >= 0 && text[k] === '\\') { backslashCount++; k--; }
                     if (backslashCount % 2 !== 0) { isDollarEscaped = true; }
                }

                if (!isDollarEscaped) {
                    const remainingText = text.substring(dollarIndex);
                    const match = remainingText.match(/^\$([\w-]+)\[/);
                    if (match) {
                        funcStartIndex = dollarIndex;
                        break;
                    }
                }
                searchPos = dollarIndex + 1;
            }


            if (funcStartIndex !== -1) {
                 const match = text.substring(funcStartIndex).match(/^\$([\w-]+)\[/);
                 const funcName = match[1];
                 const argsStartIndex = funcStartIndex + match[0].length;

                 if (funcStartIndex > currentPos) {
                     result += this.unescapeText(text.substring(currentPos, funcStartIndex));
                 }

                 const argsEndIndex = this.findMatchingBracket(text, argsStartIndex);

                 if (argsEndIndex === -1) {
                     const contextSnippet = text.substring(funcStartIndex, Math.min(funcStartIndex + 50, text.length));
                     throw new Error(`Parser Error: Unmatched bracket '[' for function "$${funcName}" starting near: "${contextSnippet}..."`);
                 }

                 const argsString = text.substring(argsStartIndex, argsEndIndex);
                 const args = await this.parseArguments(argsString, context);

                 if (context.stopExecution) break;

                 const funcResult = await this.functions.execute(funcName, args, context);

                 if (funcResult !== undefined && funcResult !== null) {
                     result += String(funcResult);
                 }
                 currentPos = argsEndIndex + 1;
                 continue;
            } else {
                 if (currentPos < text.length) {
                      result += this.unescapeText(text.substring(currentPos));
                 }
                 break;
             }
        }
        return result;
    }

    findMatchingBracket(text, startIndex) {
        let depth = 1;
        let isEscaped = false;

        for (let i = startIndex; i < text.length; i++) {
            const char = text[i];

            if (isEscaped) {
                isEscaped = false;
                continue;
            }

            if (char === '\\') {
                isEscaped = true;
                continue;
            }

            if (char === '[') {
                let j = i - 1;
                while (j >= 0 && text[j].match(/[\w-]/)) {
                    j--;
                }
                if (j >= 0 && text[j] === '$') {
                    let isDollarEscaped = false;
                    if (j > 0 && text[j - 1] === '\\') {
                         let backslashCount = 0;
                         let k = j - 1;
                         while (k >= 0 && text[k] === '\\') {
                             backslashCount++;
                             k--;
                         }
                         if (backslashCount % 2 !== 0) {
                            isDollarEscaped = true;
                         }
                    }
                    if (!isDollarEscaped) {
                        depth++;
                    }
                }
            } else if (char === ']') {
                depth--;
                if (depth === 0) {
                    return i;
                }
            }
            if (depth < 0) {
                return -1; // Should not happen with correct nesting
            }
        }
        return -1;
    }


    async parseArguments(argsString, context) {
        const args = [];
        let depth = 0;
        let currentArgStart = 0;
        let isEscaped = false;

        for (let i = 0; i < argsString.length; i++) {
             if (context.stopExecution) break;
             const char = argsString[i];

             if (isEscaped) {
                 isEscaped = false;
             } else if (char === '\\') {
                 isEscaped = true;
                 continue;
             }

             if (char === '[' && !isEscaped) {
                 let j = i - 1;
                 while (j >= 0 && argsString[j].match(/[\w-]/)) {
                     j--;
                 }
                 if (j >= 0 && argsString[j] === '$') {
                     let isDollarEscaped = false;
                     if (j > 0 && argsString[j - 1] === '\\') {
                          let backslashCount = 0;
                          let k = j - 1;
                          while (k >= 0 && argsString[k] === '\\') { backslashCount++; k--; }
                          if (backslashCount % 2 !== 0) { isDollarEscaped = true; }
                     }
                     if (!isDollarEscaped) {
                         depth++;
                     }
                 }
             } else if (char === ']' && !isEscaped) {
                  if (depth > 0) {
                      depth--;
                  }
             } else if (char === ';' && depth === 0 && !isEscaped) {
                 const argToEvaluate = argsString.substring(currentArgStart, i);
                 args.push(await this.evaluate(argToEvaluate, context));
                 currentArgStart = i + 1;
             }
        }

        if (!context.stopExecution && currentArgStart <= argsString.length) {
             const lastArgToEvaluate = argsString.substring(currentArgStart);
             args.push(await this.evaluate(lastArgToEvaluate, context));
        }

         if (argsString.length === 0 && args.length === 0) {
             args.push('');
         }
        return args;
    }

     unescapeText(text) {
         return text.replace(/\\\\/g, '\u0001')
                    .replace(/\\([$[\];])/g, '$1')
                    .replace(/\u0001/g, '\\');
     }

     async sendErrorReply(error, context) {
         const errorMessage = `❌ Error: ${error.message || 'An unknown error occurred.'}`;
         try {
             if (context.interaction && typeof context.interaction.reply === 'function') {
                 if (context.interaction.replied || context.interaction.deferred) {
                      if (typeof context.interaction.followUp === 'function') {
                         await context.interaction.followUp({ content: errorMessage, ephemeral: true });
                      }
                 } else {
                     await context.interaction.reply({ content: errorMessage, ephemeral: true });
                 }
             } else if (context.message?.channel && !error.message?.includes("reply")) {
                 await context.message.channel.send(errorMessage);
             }
         } catch (discordError) {
             if (!discordError.message?.toLowerCase().includes('unknown interaction') &&
                 !discordError.message?.toLowerCase().includes('unknown message')) {
                  console.error("[Relay.js] Failed to send error message to Discord:", discordError);
             }
         }
     }
}

module.exports = Parser;